import boto3
from .util import print_err 
import time
import os
from asgiref.sync import sync_to_async


ACCESS_KEY = os.environ.get('BEDROCK_ACCESS_KEY')
SECRET_KEY = os.environ.get('BEDROCK_SECRET_KEY')

LLAMA70B = 'us.meta.llama3-1-70b-instruct-v1:0'
LLAMA3B = 'us.meta.llama3-2-3b-instruct-v1:0'
LLAMA8B = 'us.meta.llama3-1-8b-instruct-v1:0'

MAX_ATTEMPTS = 3

def format_messages(messages):
    # Change from {'role': ROLE, 'content': MESSAGE}
    # to {'role': ROLE, 'content': [{'text': MESSAGE}]}
    for m in messages:
        if isinstance(m['content'], str):
            m['content'] = [{'text': m['content']}]
        if m['role'] == 'system':
            m['role'] = 'user'
    formatted = []
    last_msg = {}
    for m in messages:
        if last_msg.get('role') == m['role']:
            last_msg['content'].extend(m['content'])
        else:
            formatted.append(m)
            last_msg = m
    return formatted

def setup_client():
    if ACCESS_KEY and SECRET_KEY:
        return boto3.client(
            "bedrock-runtime", 
            region_name="us-east-1", 
            aws_access_key_id=ACCESS_KEY,
            aws_secret_access_key=SECRET_KEY
        )
    else: 
        print_err('BEDROCK_ACCESS_KEY, BEDROCK_SECRET_KEY keys not found on environ. Bedrock API calls will fail')

client = setup_client()

@sync_to_async
def get_response(messages, model, temp):
 return client.converse(
        modelId=model,
        messages=format_messages(messages),
        inferenceConfig={"maxTokens":4096,"stopSequences":["User:"],"temperature":temp,"topP":1},
        additionalModelRequestFields={}
    )
        
async def generate_chat_response_bedrock(messages, model, temp, attempt=0):
    try:
        response = await get_response(messages, model, temp)
        return get_message_text(response), get_cost(model, response)
      
    except (Exception) as e:
        attempt += 1
        if attempt > MAX_ATTEMPTS:
            print_err(f"ERROR: Can't invoke '{model}'. Reason: {e}")
            raise e
        sleeptime = pow(2, attempt + 1)

        time.sleep(sleeptime)
        return await generate_chat_response_bedrock(messages, model, attempt=attempt, temp=temp)

def get_message_text(response):
    return response["output"]["message"]["content"][0]["text"]

def get_tokens_used(response):
    return response['usage']['totalTokens']

def get_cost(model, response):
    token_cost = {
        LLAMA3B: 0.00015,
        LLAMA8B: 0.00022,
        LLAMA70B: 0.00099,
    }
    return token_cost.get(model, -1) * get_tokens_used(response) / 1000