import os
import json
import time
from .bedrock import generate_chat_response_bedrock
from .ollama import generate_chat_response_ollama
from .util import print_err

from openai import AsyncOpenAI
from openai import _exceptions as error


key = os.environ.get("OPENAI_KEY")

client = AsyncOpenAI(api_key=key) if key is not None else None

MAX_ATTEMPTS = 3
DEFAULT_MODEL = "gpt-4o-mini"

class InorganicFinishReasonError(Exception):
    pass

class SchemaValidationError(Exception):
    pass


# Main interface are these prompting strats


async def basic_prompt(prompt, input, model, temp):
    messages = [{"role": "user", "content": f"{prompt}\n{input}"}]
    text, cost = await generate_chat_response(messages, model, temp)
    return text, cost


async def system_prompt(prompt, input, model, temp):
    messages = [
        {"role": "system", "content": prompt},
        {"role": "user", "content": input},
    ]
    text, cost = await generate_chat_response(messages, model, temp)
    return text, cost


async def json_prompt(
    prompt_text,
    input,
    method,
    model,
    schema=None,
    attempt=1,
    best=(0, None),
    temp=1,
):
    response_txt, cost = await method(prompt_text, input, model, temp=temp)
    response_txt = response_txt.replace("```json", "").replace("```", "").strip()
    try:
        serialized = json.loads(response_txt)
    except json.decoder.JSONDecodeError as e:
        if attempt < MAX_ATTEMPTS:
            print_err(f"JSON decode error, retrying ({attempt}) {response_txt}")
            return await json_prompt(
                prompt_text, input, method, model, schema, attempt + 1, best, temp=temp
            )
        else:
            raise e
    # Schema is a method that returns a value between 0 and 1
    # 1 indicates full compliance with the schema
    # 0 indicates non compliance with the schema
    # In between values indicate partial compliance
    # The code will retry until the schema check returns 1, saving the "best so far"
    # If MAX ATTEMPTS is reached, then we return the "best so far"
    # If it's None then raise an exception
    # TBD if we should return None instead, could be useful either way
    if schema:
        check = schema(serialized)
        if check != 1:
            if attempt < MAX_ATTEMPTS:
                print_err(f"Schema check failed, retrying ({attempt}) ({serialized})")
                if check > best[0]:
                    best = (check, serialized)
                return await json_prompt(
                    prompt_text, input, method, model, schema, attempt + 1, best
                )
            else:
                if best[1]:
                    return best[1]
                else:
                    raise SchemaValidationError
    return serialized, cost


async def json_prompt_system(
    prompt_text, input, model, schema=None, temp=1
):
    return await json_prompt(
        prompt_text, input, method=system_prompt, model=model, schema=schema, temp=temp
    )

async def generate_chat_response(messages, model, temp):
    if model.startswith('gpt'):
        return await generate_chat_response_openai(messages, model, temp=temp)
    if model.startswith('ollama-'):
        model = model.removeprefix('ollama-')
        return await generate_chat_response_ollama(messages, model, temp=temp)
    if model.startswith('bedrock-'):
        model = model.removeprefix('bedrock-')
        return await generate_chat_response_bedrock(messages, model, temp=temp)
    print_err(f'Unknown model {model}. Must start with gpt, ollama-, or bedrock-')
    raise Exception

async def generate_chat_response_openai(messages, model, temp, attempt=0):
    try:
        response = await client.chat.completions.create(
            model=model, temperature=temp, messages=messages
        )

        fr = response.choices[0].finish_reason
        if fr != 'stop':
            raise InorganicFinishReasonError(fr)
        text = get_message_text(response)
        cost = get_total_cost([response])
        return text, cost
    except (
        error.InternalServerError,
        error.APIConnectionError,
        error.APITimeoutError,
        error.UnprocessableEntityError,
        error.RateLimitError,
    ) as e:
        attempt += 1
        if attempt > MAX_ATTEMPTS:
            raise e
        sleeptime = pow(2, attempt + 1)
        #print_err("Service Unavailable.. sleeping for " + str(sleeptime))
        time.sleep(sleeptime)
        return await generate_chat_response_openai(messages, model, attempt=attempt, temp=temp)
    except error.APIError as e:
        if "bad gateway" in str(e).lower():
            # bad gateway, we retry
            attempt += 1
            if attempt > MAX_ATTEMPTS:
                raise e
            sleeptime = pow(2, attempt + 1)
            print_err("Bad Gateway. Sleeping for {sleeptime}s.")
            time.sleep(sleeptime)
            return await generate_chat_response_openai(messages, model, attempt=attempt, temp=temp)
        else:
            raise e


def get_message_text(response):
    return response.choices[0].message.content


def get_total_tokens(responses):
    return sum([r.usage.total_tokens for r in responses])


def get_total_cost(responses):
    input_costs = {
        "gpt-4o-2024-08-06": 0.005 / 1000,
        "gpt-4o-2024-05-13": 0.005 / 1000,
        "gpt-4-0125-preview": 0.03 / 1000,
        "gpt-4-0613": 0.03 / 1000,
        "gpt-4-1106-preview": 0.01 / 1000,
        "gpt-4o-mini-2024-07-18": 0.15 / 1000000,
        "gpt-3.5-turbo-0613": 0.0015 / 1000,
        "gpt-3.5-turbo-0125": 0.0005 / 1000,
        "gpt-3.5-turbo-16k-0613": 0.003 / 1000,
    }
    output_costs = {
        "gpt-4o-2024-08-06": 0.015 / 1000,
        "gpt-4-0125-preview": 0.06 / 1000,
        "gpt-4-0613": 0.06 / 1000,
        "gpt-4-1106-preview": 0.03 / 1000,
        "gpt-4o-mini-2024-07-18": 0.6 / 1000000,
        "gpt-3.5-turbo-0613": 0.002 / 1000,
        "gpt-3.5-turbo-0125": 0.0015 / 1000,
        "gpt-3.5-turbo-16k-0613": 0.004 / 1000,
    }
    # check if models are present
    for r in responses:
        if r.model not in input_costs:
            print_err(f"[warning] {r.model} not found in input costs! update please")
            return -1

    return sum(
        [
            r.usage.prompt_tokens * input_costs[r.model]
            + r.usage.completion_tokens * output_costs[r.model]
            for r in responses
        ]
    )
